#!/usr/bin/env python

# ISeqDb - Identify Sequences in Databases
# version 0.0.6
# main module
# Nico Salmaso, FEM, nico.salmaso@fmach.it
# This program is distributed under the GNU General Public License (GNU GPL v.3) https://www.gnu.org/licenses/

import argparse
import sys
import textwrap
import os

from ISeqDb import find_seqs, inspect_db, create_db


def main():
    version_isdb = "0.0.6"
    description = ("ISeqDb v." + version_isdb)

    def check_path_file(filepathname):
        if not os.path.exists(filepathname):
            raise argparse.ArgumentTypeError(f"The file {filepathname} does not exist; "
                                             f"both path and filename must be indicated, "
                                             f"e.g. /path/file.fna, /path/database.tar.gz ")
        return filepathname

    parser = argparse.ArgumentParser(prog='ISeqDb', description=description,
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     epilog=textwrap.dedent('''\
                This program is distributed under the GNU General Public License (GNU GPL v.3) https://www.gnu.org/licenses/
                
                Author: Nico Salmaso (nico.salmaso@fmach.it)
                
                ISeqDb relies on BLAST®:
                
                -   BLAST® Command Line Applications User Manual [Internet]. Bethesda (MD): National Center
                    for Biotechnology Information (US); 2008-.
                -   Camacho C., Coulouris G., Avagyan V., Ma N., Papadopoulos J., Bealer K., Madden T.L. (2008)
                    “BLAST+: architecture and applications.” BMC Bioinformatics 10:421. PubMed
                ---
                ''')
                                     )
    parser.add_argument('-v', '-V', '--version', action='version', version='%(prog)s 0.0.6')
    subparsers = parser.add_subparsers(dest='command', title='commands',
                                       description='Available commands - Use "ISeqDb [command] --help"'
                                                   ' for more information about a command')

    # find_seqs
    parser_find_seqs = subparsers.add_parser('find_seqs', help='Identify sequences in databases',
                                             formatter_class=argparse.RawDescriptionHelpFormatter,
                                             epilog=textwrap.dedent('''\

             Output legend (in square brackets, the blast codes)
                ---
                query_id:		        query/sequence identifier [qacc]
                subject_accession:	        NCBI accession number or subject identifier in the database [sacc]
                pident:			percentage of identical matches in query and subject sequences [pident]
                n_ident_match:		number of identical bases/matches [nident]
                n_diff_match:		number of different bases/matches [mismatch]
                n_gaps:			total number of gaps [gaps]
                alignment_length:	        length of the alignemnt between query and subject sequences [length]
                query_start_al:		start of alignment in query [qstart]
                query_end_al:		end of alignment in query [qend]
                subj_start_al:		start of alignment in subject [sstart]
                subj_end_al:		        end of alignment in subject [send]
                qlen:                        query sequence length
                slen:                        subject sequence length
                bitscore:			bit score [bitscore]
                evalue:			expect value [evalue]
                subject_seq_title:		title of subject sequence in database [stitle]
                align_query_seq:		aligned part of query sequence [qseq]
                align_subj_seq:		aligned part of subject sequence [sseq]
                ---
             ''')
                                             )
    parser_find_seqs.add_argument('queryfile', type=check_path_file, default=None,
                                  help="path/file with extension .fasta .fas .fa .fna .ffn .faa; e.g. /dir/query.fasta")
    parser_find_seqs.add_argument('targetdb', type=check_path_file, default=None,
                                  help="Database for the blast analysis (with path) - e.g. /dir/arch.tar.gz")
    parser_find_seqs.add_argument('outputfile', default=None,
                                  help="Output name file (with path) - e.g. /dir/out.txt")
    parser_find_seqs.add_argument('-k', '--task', type=str, required=False, default="megablast",
                                  choices=['megablast', 'blastn', 'blastx'],
                                  help="Task: megablast (nucl), blastn (nucl), blastx (prot); default=megablast")
    parser_find_seqs.add_argument('-m', '--maxtargseq', type=int, required=False, default=50,
                                  help="Keep max target sequences >= maxtargseq; default=50")
    parser_find_seqs.add_argument('-e', '--minevalue', type=float, required=False, default=1e-6,
                                  help="Keep hits with evalue <= minevalue; default=1e-6")
    parser_find_seqs.add_argument('-p', '--minpident', type=int, required=False, default=85,
                                  help="Keep hits with pident >= minpident (only megablast/blastn); default=85")
    parser_find_seqs.add_argument('-t', '--threads', type=int, required=False, default=1,
                                  help="Number of threads to use; default=1")
    parser_find_seqs.add_argument('-s', '--sortoutput', type=str, required=False, default="bitscore",
                                  choices=['bitscore', 'pident', 'evalue', 'subject_seq_title'],
                                  help="Sort output by colname; default=bitscore")
    parser_find_seqs.add_argument('-d', '--delimiter', type=str, required=False, default="tab",
                                  choices=['comma', 'semicolon', 'tab'],
                                  help="Output delimiter: comma, semicolon, tab; default=tab")
    parser_find_seqs.set_defaults(func=find_seqs.run)

    # inspect_db
    parser_inspect_db = subparsers.add_parser('inspect_db', help='Inspect or save the sequence database')
    parser_inspect_db.add_argument('targetdb_inspect', type=check_path_file, default=None,
                                   help="Database for the blast analysis (with path) - e.g. /dir/arch.tar.gz")
    parser_inspect_db.add_argument('-d', '--savedb', required=False, default="nosave",
                                   help="Output directory (with path) - e.g. /dir/; if not indicated, default=nosave")
    parser_inspect_db.set_defaults(func=inspect_db.run)

    # create_db
    parser_create_db = subparsers.add_parser('create_db', help='Create a database from a fasta/multifasta file')
    parser_create_db.add_argument('targetfasta', type=check_path_file, default=None,
                                  help="path/file with extension .fasta .fas .fa .fna .ffn .faa); e.g. /dir/arch.fasta")
    parser_create_db.add_argument('-m', '--moltype', type=str, required=True, default=None, choices=['nucl', 'prot'],
                                  help="Molecule type in fasta file, nucl (nucleotide) or prot (protein)")
    parser_create_db.set_defaults(func=create_db.run)

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



# ISeqDb - Identify Sequences in Databases
# version 0.0.6
# module, create_db
# Nico Salmaso, FEM, nico.salmaso@fmach.it
# This program is distributed under the GNU General Public License (GNU GPL v.3) https://www.gnu.org/licenses/

import os
import sys
import tarfile


def run(args):

    print('_____________________________________')
    print(' ')
    print("Database:", args.targetfasta)
    print('_____________________________________')
    print(' ')

    dirfasta, file_name_fs = os.path.split(args.targetfasta)
    datseq = file_name_fs.split('.')[0]

    if len(dirfasta) == 0:
        print(" ")
        print("input directory not indicated; full path should be indicated")
        print(" ")
        sys.exit(1)
    if len(file_name_fs) == 0:
        print(" ")
        print("fasta file name not indicated")
        print(" ")
        sys.exit(1)
    if any(c in " ;&|$-" for c in dirfasta):
        print(" ")
        print(dirfasta, file_name_fs)
        print("spaces or characters ;&|$- in input dir name are not allowed")
        print(" ")
        sys.exit(1)
    if any(c in " ;&|$-" for c in file_name_fs):
        print(" ")
        print(dirfasta, file_name_fs)
        print("spaces or characters ;&|$- in input file name are not allowed")
        print(" ")
        sys.exit(1)

    # ##############################################
    # check if the query file has the correct extensions
    file_exte = ['.fasta', '.fas', '.fa', '.fna', '.ffn', '.faa']
    exte = os.path.splitext(args.targetfasta)[1]
    if exte.lower() in file_exte:
        print("fasta ext pass")
    else:
        print(f"{args.targetfasta} does not have a valid file extension .fasta .fas .fa .fna .ffn .faa")
        sys.exit(1)

    # check if the query file begins with > (greater-than)
    def check_char_1(filename):
        try:
            with open(filename, 'r') as file:
                char_1 = file.read(1)
                return char_1 == '>'
        except FileNotFoundError:
            print(f"{filename} does not exist")
            return False
        except UnicodeDecodeError:
            print("")
            print(f"{filename} is binary or contains invalid data")
            sys.exit(1)

    if check_char_1(args.targetfasta):
        print("ch_1 pass")
    else:
        print("")
        print(f"{args.targetfasta}: line 1 is not compatible with the structure of a fasta file")
        sys.exit(1)

    # ##############################################

    outdir = os.path.join(str(dirfasta), str(datseq) + '_out')

    if os.path.isdir(outdir):
        print("Cannot write an existing directory: ", outdir)
        print(" ")
        sys.exit(1)
    else:
        os.mkdir(outdir)

    define_cmd_1: str = (
        f'makeblastdb -in {args.targetfasta} -input_type fasta -dbtype {args.moltype} '
        f'-parse_seqids -out {dirfasta}"/"{datseq} '
    )
    os.system(define_cmd_1)

    if args.moltype == "nucl":
        chartask = ".n"
    elif args.moltype == "prot":
        chartask = ".p"
    else:
        print(" ")
        print("allowed moltype are nucl (nucleotides) or prot (proteins)")
        sys.exit(1)

    elements_db = [
        ("".join([datseq, chartask, 'db'])),
        ("".join([datseq, chartask, 'hr'])),
        ("".join([datseq, chartask, 'in'])),
        ("".join([datseq, chartask, 'js'])),
        ("".join([datseq, chartask, 'og'])),
        ("".join([datseq, chartask, 'os'])),
        ("".join([datseq, chartask, 'ot'])),
        ("".join([datseq, chartask, 'sq'])),
        ("".join([datseq, chartask, 'tf'])),
        ("".join([datseq, chartask, 'to']))
    ]

    for fileelem in elements_db:
        os.rename(os.path.join(dirfasta, fileelem), os.path.join(outdir, fileelem))

    filetargz = (datseq + '.tar.gz')
    output_targz = os.path.join(dirfasta, filetargz)

    with tarfile.open(output_targz, "w:gz") as tar:
        for fileelem in elements_db:
            direl = os.path.join(outdir, fileelem)
            tar.add(direl, arcname=fileelem)
            os.remove(direl)

    os.rmdir(outdir)



# ISeqDb - Identify Sequences in Databases
# version 0.0.6
# module find_seqs
# Nico Salmaso, FEM, nico.salmaso@fmach.it
# This program is distributed under the GNU General Public License (GNU GPL v.3) https://www.gnu.org/licenses/

import pandas as pd
import os
import sys
import tarfile
from pandas import DataFrame


def run(args):

    dirquery, file_query = os.path.split(args.queryfile)
    if len(dirquery) == 0:
        print(" ")
        print("query directory not indicated")
        print(" ")
        sys.exit(1)
    if len(file_query) == 0:
        print(" ")
        print("name query file not indicated")
        print(" ")
        sys.exit(1)
    if any(c in " ;&|$-" for c in dirquery):
        print(" ")
        print(dirquery, file_query)
        print("spaces or characters ;&|$- in input dir name are not allowed")
        print(" ")
        sys.exit(1)
    if any(c in " ;&|$-" for c in file_query):
        print(" ")
        print(dirquery, file_query)
        print("spaces or characters ;&|$- in input file name are not allowed")
        print(" ")
        sys.exit(1)

    dirtarget, file_name_db = os.path.split(args.targetdb)
    if len(dirtarget) == 0:
        print(" ")
        print("db directory not indicated")
        print(" ")
        sys.exit(1)
    if len(file_name_db) == 0:
        print(" ")
        print("database.tar.gz name not indicated")
        print(" ")
        sys.exit(1)
    if any(c in " ;&|$-" for c in dirtarget):
        print(" ")
        print(dirtarget, file_name_db)
        print("spaces or characters ;&|$- in input dir name are not allowed")
        print(" ")
        sys.exit(1)
    if any(c in " ;&|$-" for c in file_name_db):
        print(" ")
        print(dirtarget, file_name_db)
        print("spaces or characters ;&|$- in input file name are not allowed")
        print(" ")
        sys.exit(1)

    diroutput, file_name_ou_1 = os.path.split(args.outputfile)
    if len(diroutput) == 0:
        print(" ")
        print("output directory not indicated")
        print(" ")
        sys.exit(1)
    if len(file_name_ou_1) == 0:
        print(" ")
        print("output file name not indicated")
        print(" ")
        sys.exit(1)
    if any(c in " ;&|$-" for c in diroutput):
        print(" ")
        print(diroutput, file_name_ou_1)
        print("spaces or characters ;&|$- in input dir name are not allowed")
        print(" ")
        sys.exit(1)
    if any(c in " ;&|$-" for c in file_name_ou_1):
        print(" ")
        print(diroutput, file_name_ou_1)
        print("spaces or characters ;&|$- in input file name are not allowed")
        print(" ")
        sys.exit(1)

    # ##############################################

    # check if the query file has the correct extensions
    file_exte = ['.fasta', '.fas', '.fa', '.fna', '.ffn', '.faa']
    exte = os.path.splitext(args.queryfile)[1]
    if exte.lower() in file_exte:
        print("fasta ext pass")
    else:
        print(f"{args.queryfile} does not have a valid file extension .fasta .fas .fa .fna .ffn .faa")
        sys.exit(1)

    # check if the subject database has the correct extension
    if args.targetdb.endswith('.tar.gz'):
        print("db ext pass")
    else:
        print(f"{args.targetdb} does not have a valid file extension .tar.gz")
        sys.exit(1)

    # check if the query file begins with > (greater-than)
    def check_char_1(filename):
        try:
            with open(filename, 'r') as file:
                char_1 = file.read(1)
                return char_1 == '>'
        except FileNotFoundError:
            print(f"{filename} does not exist")
            return False
        except UnicodeDecodeError:
            print("")
            print(f"{filename} is binary or contains invalid data")
            sys.exit(1)

    if check_char_1(args.queryfile):
        print("ch_1 pass")
    else:
        print("")
        print(f"{args.queryfile}: line 1 is not compatible with the structure of a fasta file")
        sys.exit(1)

    # ##############################################

    # Other check

    if not (1 <= args.minpident <= 100):
        print(f"Error: pident {args.minpident} is not in the range 1:100")
        sys.exit(1)

    # ##############################################

    file_name_ou = file_name_ou_1.split('.')[0]

    # open and extract database
    datseqtar = tarfile.open(args.targetdb)
    # extracting file
    datseqtar.extractall(dirtarget)
    datseqtar.close()

    datseq = args.targetdb.split('.')[0]

    if (args.task == "megablast") or (args.task == "blastn"):
        define_cmd: str = (
            f'blastn -query {args.queryfile} -db {datseq} -task {args.task} '
            f'-perc_identity {args.minpident} -dust no -evalue {args.minevalue} '
            f'-outfmt \'6 delim=  qacc sacc pident nident mismatch gaps length '
            f'qstart qend sstart send qlen slen bitscore evalue stitle qseq sseq\' '
            f'-max_target_seqs {args.maxtargseq} -num_threads {args.threads} '
            f'| tee -i {args.outputfile}_nh.txt'
        )
        os.system(define_cmd)
    elif args.task == "blastx":
        define_cmd: str = (
            f'blastx -query {args.queryfile} -db {datseq} -task {args.task} '
            f'-evalue {args.minevalue} '
            f'-outfmt \'6 delim=  qacc sacc pident nident mismatch gaps length '
            f'qstart qend sstart send qlen slen bitscore evalue stitle qseq sseq\' '
            f'-max_target_seqs {args.maxtargseq} -num_threads {args.threads} '
            f'| tee -i {args.outputfile}_nh.txt'
        )
        os.system(define_cmd)
    else:
        print(" ")
        print("allowed task: megablast, blastn, blastx")
        sys.exit(1)

    # cleaning 1

    if (args.task == "megablast") or (args.task == "blastn"):
        chartask = ".n"
    elif args.task == "blastx":
        chartask = ".p"
    else:
        print(" ")
        print("allowed tasks are megablast, blastn or blastx")
        sys.exit(1)

    os.remove("".join([datseq, chartask, 'db']))
    os.remove("".join([datseq, chartask, 'hr']))
    os.remove("".join([datseq, chartask, 'in']))
    os.remove("".join([datseq, chartask, 'js']))
    os.remove("".join([datseq, chartask, 'og']))
    os.remove("".join([datseq, chartask, 'os']))
    os.remove("".join([datseq, chartask, 'ot']))
    os.remove("".join([datseq, chartask, 'sq']))
    os.remove("".join([datseq, chartask, 'tf']))
    os.remove("".join([datseq, chartask, 'to']))

    column_names: list[str] = ['query_id', 'subject_accession', 'pident', 'n_ident_match', 'n_diff_match',
                               'n_gaps', 'alignment_length', 'query_start_al', 'query_end_al', 'subj_start_al',
                               'subj_end_al', 'qlen', 'slen', 'bitscore', 'evalue', 'subject_seq_title',
                               'align_query_seq', 'align_subj_seq']

    file_header = f"{args.outputfile}_nh.txt"
    output_file = f"{args.outputfile}"

    # when the file is empty (no matches)
    if os.stat(file_header).st_size == 0:
        with open(file_header, 'w') as f:
            column_names_str = '\t'.join(map(str, column_names))
            f.write(column_names_str)
            print("")
            print('_____________________________________')
            print("")
            print("No matches between query and subject sequences")
            print('_____________________________________')

    # Read the file into a DataFrame; do not consider the first line in the form of header
    dframe = pd.read_csv(file_header, delimiter='\t', header=None)
    # Convert the dframe (list of lists) and add column names in the first row
    data = dframe.values.tolist()
    data.insert(0, column_names)
    dataf: DataFrame = pd.DataFrame(data)
    dataf.to_csv(output_file, sep='\t', index=False, header=False)

    # sort file
    outputfile = pd.read_csv(args.outputfile, sep='\t')
    outputfile_sorted = outputfile.sort_values(by=args.sortoutput, ascending=False)

    if args.delimiter == "semicolon":
        sorted_file_name = os.path.join(diroutput, file_name_ou + '_sorted.csv2')
        sorted_file_name = str(sorted_file_name)
        outputfile_sorted.to_csv(sorted_file_name, sep=";", index=False)
    elif args.delimiter == "tab":
        sorted_file_name = os.path.join(diroutput, file_name_ou + '_sorted.tsv')
        sorted_file_name = str(sorted_file_name)
        outputfile_sorted.to_csv(sorted_file_name, sep="\t", index=False)
    else:
        sorted_file_name = os.path.join(diroutput, file_name_ou + '_sorted.csv')
        sorted_file_name = str(sorted_file_name)
        outputfile_sorted.to_csv(sorted_file_name, sep=",", index=False)

    print(' ')
    print('_____________________________________')
    print(' ')
    print("queryfile:", args.queryfile)
    print('')
    print("targetdb:", args.targetdb)
    print('')
    print("outputfile:", args.outputfile)
    print('')
    print("outputfile (sorted):", sorted_file_name)
    print('')
    print("minpident:", args.minpident)
    print('')
    print("task:", args.task)
    print('')
    print("minevalue:", args.minevalue)
    print('_____________________________________')
    print(' ')

    if args.task == "blastx":
        print('')
        print("minpident not defined in blastx")

    # cleaning 2
    os.remove(args.outputfile + '_nh.txt')



# ISeqDb - Identify Sequences in Databases
# version 0.0.6
# module inspect_db
# Nico Salmaso, FEM, nico.salmaso@fmach.it
# This program is distributed under the GNU General Public License (GNU GPL v.3) https://www.gnu.org/licenses/

import os
import tarfile
import sys


def run(args):

    # ##############################################
    # check if the database has the correct extension
    if args.targetdb_inspect.endswith('.tar.gz'):
        print("db ext pass")
    else:
        print(f"{args.targetdb_inspect} does not have a valid file extension .tar.gz")
        sys.exit(1)
    # ##############################################

    print('_____________________________________')
    print(' ')
    print("Database:", args.targetdb_inspect)
    print('')
    print('_____________________________________')
    print(' ')

    dirtarget, file_name_db = os.path.split(args.targetdb_inspect)
    if len(dirtarget) == 0:
        print(" ")
        print("db directory not indicated")
        print(" ")
        sys.exit(1)
    if len(file_name_db) == 0:
        print(" ")
        print("database.tar.gz name not indicated")
        print(" ")
        sys.exit(1)
    if any(c in " ;&|$-" for c in dirtarget):
        print(" ")
        print(dirtarget, file_name_db)
        print("spaces or characters ;&|$- in input dir name are not allowed")
        print(" ")
        sys.exit(1)
    if any(c in " ;&|$-" for c in file_name_db):
        print(" ")
        print(dirtarget, file_name_db)
        print("spaces or characters ;&|$- in input file name are not allowed")
        print(" ")
        sys.exit(1)

    # open and extract database
    datseqtar = tarfile.open(args.targetdb_inspect)
    # extracting file
    datseqtar.extractall(dirtarget)
    datseqtar.close()

    # remove extensions
    datseq = args.targetdb_inspect.split('.')[0]
    datseq_name = file_name_db.split('.')[0]

    define_cmd_1: str = (
        f'blastdbcmd -db {datseq} -entry all -outfmt %f'
    )
    os.system(define_cmd_1)

    define_cmd_2: str = (
        f'blastdbcmd -db {datseq} -info'
    )
    os.system(define_cmd_2)

    if args.savedb != "nosave":
        file_name = os.path.join(args.savedb, datseq_name + '.fasta')
        file_name = str(file_name)
        define_cmd_3: str = (
            f'blastdbcmd -db {datseq} -entry all -outfmt %f > {file_name}'
        )
        os.system(define_cmd_3)
        print(" ")
        print("db saved: ", file_name)
        print(" ")
    else:
        print(" ")
        print("db not saved")
        print(" ")

    # cleaning
    if os.path.isfile(datseq + '.njs'):
        chartask = ".n"
    elif os.path.isfile(datseq + '.pjs'):
        chartask = ".p"
    else:
        print(" ")
        print("unidentified molecule")
        sys.exit(1)
    os.remove("".join([datseq, chartask, 'db']))
    os.remove("".join([datseq, chartask, 'hr']))
    os.remove("".join([datseq, chartask, 'in']))
    os.remove("".join([datseq, chartask, 'js']))
    os.remove("".join([datseq, chartask, 'og']))
    os.remove("".join([datseq, chartask, 'os']))
    os.remove("".join([datseq, chartask, 'ot']))
    os.remove("".join([datseq, chartask, 'sq']))
    os.remove("".join([datseq, chartask, 'tf']))
    os.remove("".join([datseq, chartask, 'to']))
